from mlhandmade.discriminant_analysis.lda import LinearDiscriminantAnalysis
from mlhandmade.discriminant_analysis.qda import QuadraticDiscriminantAnalysis

__all__ = [
    "LinearDiscriminantAnalysis",
    "QuadraticDiscriminantAnalysis"
]